@@ -2,15 +2,18 @@
 from __future__ import annotations
 
 import os
+import sys
 from distutils.spawn import find_executable
-from typing import Any, List, Optional, Union
+from typing import Any, List, Optional, Union, TextIO
 from typing_extensions import Literal
 
 import argo_client.interaction as argo
 from argo_client.connection import DynamicSocketProcess, ServerConnection, ServerProcess, StdIOProcess, HttpProcess
+from .custom_fstring import *
+from .quoting import *
 from . import cryptoltypes
 from . import solver
-from .commands import * 
+from .commands import *
 
 
 def connect(command : Optional[str]=None,
@@ -18,7 +21,9 @@ def connect(command : Optional[str]=None,
             cryptol_path : Optional[str] = None,
             url : Optional[str] = None,
             reset_server : bool = False,
-            verify : Union[bool, str] = True) -> CryptolConnection:
+            verify : Union[bool, str] = True,
+            log_dest : Optional[TextIO] = None,
+            timeout : Optional[float] = None) -> CryptolConnection:
     """
     Connect to a (possibly new) Cryptol server process.
 
@@ -39,6 +44,14 @@ def connect(command : Optional[str]=None,
                    only has an affect when ``connect`` is called with a ``url`` parameter
                    or when the ``CRYPTOL_SERVER_URL`` environment variable is set.
 
+    :param log_dest: A destination to log JSON requests/responses to, e.g. ``log_dest=sys.stderr``
+    will print traffic to ``stderr``, ``log_dest=open('foo.log', 'w')`` will log to ``foo.log``,
+    etc.
+
+    :param timeout: Optional default timeout (in seconds) for methods. Can be modified/read via the
+    `timeout` member field on a `CryptolConnection` or the `get_default_timeout` and
+    `set_default_timeout` methods. Method invocations which specify the optional `timeout` keyword
+    parameter will cause the default to be ignored for that method.
 
     If no ``command`` or ``url`` parameters are provided, the following are attempted in order:
 
@@ -56,27 +69,27 @@ def connect(command : Optional[str]=None,
     if command is not None:
         if url is not None:
             raise ValueError("A Cryptol server URL cannot be specified with a command currently.")
-        c = CryptolConnection(command, cryptol_path)
+        c = CryptolConnection(command, cryptol_path, log_dest=log_dest, timeout=timeout)
     # User-passed url?
     if c is None and url is not None:
-        c = CryptolConnection(ServerConnection(HttpProcess(url, verify=verify)), cryptol_path)
+        c = CryptolConnection(ServerConnection(HttpProcess(url, verify=verify)), cryptol_path, log_dest=log_dest)
     # Check `CRYPTOL_SERVER` env var if no connection identified yet
     if c is None:
         command = os.getenv('CRYPTOL_SERVER')
         if command is not None:
             command = find_executable(command)
             if command is not None:
-                c = CryptolConnection(command+" socket", cryptol_path=cryptol_path)
+                c = CryptolConnection(command+" socket", cryptol_path=cryptol_path, log_dest=log_dest, timeout=timeout)
     # Check `CRYPTOL_SERVER_URL` env var if no connection identified yet
     if c is None:
         url = os.getenv('CRYPTOL_SERVER_URL')
         if url is not None:
-            c = CryptolConnection(ServerConnection(HttpProcess(url,verify=verify)), cryptol_path)
+            c = CryptolConnection(ServerConnection(HttpProcess(url,verify=verify)), cryptol_path, log_dest=log_dest, timeout=timeout)
     # Check if `cryptol-remote-api` is in the PATH if no connection identified yet
     if c is None:
         command = find_executable('cryptol-remote-api')
         if command is not None:
-            c = CryptolConnection(command+" socket", cryptol_path=cryptol_path)
+            c = CryptolConnection(command+" socket", cryptol_path=cryptol_path, log_dest=log_dest, timeout=timeout)
     # Raise an error if still no connection identified yet
     if c is not None:
         if reset_server:
@@ -92,7 +105,10 @@ def connect(command : Optional[str]=None,
 
 
 
-def connect_stdio(command : str, cryptol_path : Optional[str] = None) -> CryptolConnection:
+def connect_stdio(command : str,
+                  cryptol_path : Optional[str] = None,
+                  log_dest : Optional[TextIO] = None,
+                  timeout : Optional[float] = None) -> CryptolConnection:
     """Start a new connection to a new Cryptol server process.
 
     :param command: The command to launch the Cryptol server.
@@ -100,10 +116,19 @@ def connect_stdio(command : str, cryptol_path : Optional[str] = None) -> Cryptol
     :param cryptol_path: An optional replacement for the contents of
       the ``CRYPTOLPATH`` environment variable.
 
+    :param log_dest: A destination to log JSON requests/responses to, e.g. ``log_dest=sys.stderr``
+    will print traffic to ``stderr``, ``log_dest=open('foo.log', 'w')`` will log to ``foo.log``,
+    etc.
+
+    :param timeout: Optional default timeout (in seconds) for methods. Can be modified/read via the
+    `timeout` member field on a `CryptolConnection` or the `get_default_timeout` and
+    `set_default_timeout` methods. Method invocations which specify the optional `timeout` keyword
+    parameter will cause the default to be ignored for that method.
+
     """
     conn = CryptolStdIOProcess(command, cryptol_path=cryptol_path)
 
-    return CryptolConnection(conn)
+    return CryptolConnection(conn, log_dest=log_dest, timeout=timeout)
 
 
 class CryptolConnection:
@@ -120,21 +145,36 @@ class CryptolConnection:
     sequential state dependencies between commands.
     """
     most_recent_result : Optional[argo.Interaction]
-    proc: ServerProcess
+
+    timeout : Optional[float]
+    """(Optional) default timeout in seconds for methods."""
 
     def __init__(self,
                 command_or_connection : Union[str, ServerConnection, ServerProcess],
-                cryptol_path : Optional[str] = None) -> None:
+                cryptol_path : Optional[str] = None,
+                *,
+                log_dest : Optional[TextIO] = None,
+                timeout : Optional[float] = None) -> None:
         self.most_recent_result = None
+        self.timeout = timeout
         if isinstance(command_or_connection, ServerProcess):
-            self.proc = command_or_connection
-            self.server_connection = ServerConnection(self.proc)
+            self.server_connection = ServerConnection(command_or_connection)
         elif isinstance(command_or_connection, str):
-            self.proc = CryptolDynamicSocketProcess(command_or_connection, cryptol_path=cryptol_path)
-            self.server_connection = ServerConnection(self.proc)
+            self.server_connection = ServerConnection(CryptolDynamicSocketProcess(command_or_connection, cryptol_path=cryptol_path))
         else:
             self.server_connection = command_or_connection
 
+        if log_dest:
+            self.logging(on=True,dest=log_dest)
+
+    def get_default_timeout(self) -> Optional[float]:
+        """Get the value of the optional default timeout for methods (in seconds)."""
+        return self.timeout
+    
+    def set_default_timeout(self, timeout : Optional[float]) -> None:
+        """Set the value of the optional default timeout for methods (in seconds)."""
+        self.timeout = timeout
+    
     # Currently disabled in (overly?) conservative effort to not accidentally
     # duplicate and share mutable state.
 
@@ -154,96 +194,204 @@ class CryptolConnection:
             return self.most_recent_result.state()
 
     # Protocol messages
-    def load_file(self, filename : str) -> argo.Command:
+    def load_file(self, filename : str, *, timeout:Optional[float] = None) -> argo.Command:
         """Load a filename as a Cryptol module, like ``:load`` at the Cryptol
         REPL.
+
+        :param timeout: Optional timeout for this request (in seconds).
         """
-        self.most_recent_result = CryptolLoadFile(self, filename)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolLoadFile(self, filename, timeout)
         return self.most_recent_result
 
-    def load_module(self, module_name : str) -> argo.Command:
-        """Load a Cryptol module, like ``:module`` at the Cryptol REPL."""
-        self.most_recent_result = CryptolLoadModule(self, module_name)
-        return self.most_recent_result
+    def load_module(self, module_name : str, *, timeout:Optional[float] = None) -> argo.Command:
+        """Load a Cryptol module, like ``:module`` at the Cryptol REPL.
 
-    def eval(self, expression : Any) -> argo.Command:
-        """Evaluate a Cryptol expression, represented according to
-        :ref:`cryptol-json-expression`, with Python datatypes standing
-        for their JSON equivalents.
+        :param timeout: Optional timeout for this request (in seconds).
         """
-        self.most_recent_result = CryptolEvalExpr(self, expression)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolLoadModule(self, module_name, timeout)
         return self.most_recent_result
 
-    def evaluate_expression(self, expression : Any) -> argo.Command:
+    def eval_raw(self, expression : Any, *, timeout:Optional[float] = None) -> argo.Command:
+        """Like the member method ``eval``, but does not call
+        ``from_cryptol_arg`` on the ``.result()``.
+
+        :param timeout: Optional timeout for this request (in seconds).
+        """
+        if hasattr(expression, '__to_cryptol__'):
+            expression = cryptoltypes.to_cryptol(expression)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolEvalExprRaw(self, expression, timeout)
+        return self.most_recent_result
+
+    def eval(self, expression : Any, *, timeout:Optional[float] = None) -> argo.Command:
+        """Evaluate a Cryptol expression, with the result represented
+        according to :ref:`cryptol-json-expression`, with Python datatypes
+        standing for their JSON equivalents.
+
+        :param timeout: Optional timeout for this request (in seconds).
+        """
+        if hasattr(expression, '__to_cryptol__'):
+            expression = cryptoltypes.to_cryptol(expression)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolEvalExpr(self, expression, timeout)
+        return self.most_recent_result
+
+    def eval_f(self, s : str, *, timeout:Optional[float] = None) -> argo.Command:
+        """Parses the given string like ``cry_f``, then evalues it, with the
+        result represented according to :ref:`cryptol-json-expression`, with
+        Python datatypes standing for their JSON equivalents.
+
+        :param timeout: Optional timeout for this request (in seconds).
+        """
+        expression = to_cryptol_str_customf(s, frames=1, filename="<eval_f>")
+        return self.eval(expression, timeout=timeout)
+
+    def evaluate_expression(self, expression : Any, *, timeout:Optional[float] = None) -> argo.Command:
         """Synonym for member method ``eval``.
+
+        :param timeout: Optional timeout for this request (in seconds)."""
+        return self.eval(expression, timeout=timeout)
+
+    def extend_search_path(self, *dir : str, timeout:Optional[float] = None) -> argo.Command:
+        """Extend the search path for loading Cryptol modules.
+
+        :param timeout: Optional timeout for this request (in seconds)."""
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolExtendSearchPath(self, list(dir), timeout)
+        return self.most_recent_result
+
+    def call_raw(self, fun : str, *args : List[Any], timeout:Optional[float] = None) -> argo.Command:
+        """Like the member method ``call``, but does not call
+        ``from_cryptol_arg`` on the ``.result()``.
         """
-        return self.eval(expression)
-
-    def extend_search_path(self, *dir : str) -> argo.Command:
-        """Load a Cryptol module, like ``:module`` at the Cryptol REPL."""
-        self.most_recent_result = CryptolExtendSearchPath(self, list(dir))
+        timeout = timeout if timeout is not None else self.timeout
+        encoded_args = [cryptoltypes.to_cryptol(a) for a in args]
+        self.most_recent_result = CryptolCallRaw(self, fun, encoded_args, timeout)
         return self.most_recent_result
 
-    def call(self, fun : str, *args : List[Any]) -> argo.Command:
-        encoded_args = [cryptoltypes.CryptolType().from_python(a) for a in args]
-        self.most_recent_result = CryptolCall(self, fun, encoded_args)
+    def call(self, fun : str, *args : List[Any], timeout:Optional[float] = None) -> argo.Command:
+        """Call function ``fun`` with specified ``args``.
+
+        :param timeout: Optional timeout for this request (in seconds)."""
+        timeout = timeout if timeout is not None else self.timeout
+        encoded_args = [cryptoltypes.to_cryptol(a) for a in args]
+        self.most_recent_result = CryptolCall(self, fun, encoded_args, timeout)
         return self.most_recent_result
 
-    def check(self, expr : Any, *, num_tests : Union[Literal['all'], int, None] = None) -> argo.Command:
+
+    def check_raw(self, expr : Any, *, num_tests : Union[Literal['all'], int, None] = None, timeout:Optional[float] = None) -> argo.Command:
+        """Like the member method ``check``, but does not call
+        `to_check_report` on the ``.result()``.
+        """
+        timeout = timeout if timeout is not None else self.timeout
+        if num_tests == "all" or isinstance(num_tests, int) or num_tests is None:
+            self.most_recent_result = CryptolCheckRaw(self, expr, num_tests, timeout)
+            return self.most_recent_result
+        else:
+            raise ValueError('``num_tests`` must be an integer, ``None``, or the string literall ``"all"``')
+
+    def check(self, expr : Any, *, num_tests : Union[Literal['all'], int, None] = None, timeout:Optional[float] = None) -> argo.Command:
         """Tests the validity of a Cryptol expression with random inputs. The expression must be a function with
         return type ``Bit``.
 
         If ``num_tests`` is ``"all"`` then the expression is tested exhaustively (i.e., against all possible inputs).
 
         If ``num_tests`` is omitted, Cryptol defaults to running 100 tests.
+
+        :param timeout: Optional timeout for this request (in seconds).
         """
+        timeout = timeout if timeout is not None else self.timeout
         if num_tests == "all" or isinstance(num_tests, int) or num_tests is None:
-            self.most_recent_result = CryptolCheck(self, expr, num_tests)
+            self.most_recent_result = CryptolCheck(self, expr, num_tests, timeout)
             return self.most_recent_result
         else:
             raise ValueError('``num_tests`` must be an integer, ``None``, or the string literall ``"all"``')
 
 
-    def check_type(self, code : Any) -> argo.Command:
+    def check_type(self, code : Any, *, timeout:Optional[float] = None) -> argo.Command:
         """Check the type of a Cryptol expression, represented according to
         :ref:`cryptol-json-expression`, with Python datatypes standing for
         their JSON equivalents.
+
+        :param timeout: Optional timeout for this request (in seconds).
         """
-        self.most_recent_result = CryptolCheckType(self, code)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolCheckType(self, code, timeout)
         return self.most_recent_result
 
-    def sat(self, expr : Any, solver : solver.Solver = solver.Z3, count : int = 1) -> argo.Command:
+    def sat_raw(self, expr : Any, solver : solver.Solver = solver.Z3, count : int = 1, *, timeout:Optional[float] = None) -> argo.Command:
+        """Like the member method ``sat``, but does not call
+        `to_smt_query_result` on the ``.result()``.
+        """
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolSatRaw(self, expr, solver, count, timeout)
+        return self.most_recent_result
+
+    def sat(self, expr : Any, solver : solver.Solver = solver.Z3, count : int = 1, *, timeout:Optional[float] = None) -> argo.Command:
         """Check the satisfiability of a Cryptol expression, represented according to
         :ref:`cryptol-json-expression`, with Python datatypes standing for
         their JSON equivalents. Use the solver named `solver`, and return up to
         `count` solutions.
+
+        :param timeout: Optional timeout for this request (in seconds).
         """
-        self.most_recent_result = CryptolSat(self, expr, solver, count)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolSat(self, expr, solver, count, timeout)
         return self.most_recent_result
 
-    def prove(self, expr : Any, solver : solver.Solver = solver.Z3) -> argo.Command:
+    def prove_raw(self, expr : Any, solver : solver.Solver = solver.Z3, *, timeout:Optional[float] = None) -> argo.Command:
+        """Like the member method ``prove``, but does not call
+        `to_smt_query_result` on the ``.result()``.
+        """
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolProveRaw(self, expr, solver, timeout)
+        return self.most_recent_result
+
+    def prove(self, expr : Any, solver : solver.Solver = solver.Z3, *, timeout:Optional[float] = None) -> argo.Command:
         """Check the validity of a Cryptol expression, represented according to
         :ref:`cryptol-json-expression`, with Python datatypes standing for
         their JSON equivalents. Use the solver named `solver`.
+
+        :param timeout: Optional timeout for this request (in seconds).
         """
-        self.most_recent_result = CryptolProve(self, expr, solver)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolProve(self, expr, solver, timeout)
         return self.most_recent_result
 
-    def safe(self, expr : Any, solver : solver.Solver = solver.Z3) -> argo.Command:
+    def safe_raw(self, expr : Any, solver : solver.Solver = solver.Z3, *, timeout:Optional[float] = None) -> argo.Command:
+        """Like the member method ``safe``, but does not call
+        `to_smt_query_result` on the ``.result()``.
+        """
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolSafeRaw(self, expr, solver, timeout)
+        return self.most_recent_result
+
+    def safe(self, expr : Any, solver : solver.Solver = solver.Z3, *, timeout:Optional[float] = None) -> argo.Command:
         """Check via an external SMT solver that the given term is safe for all inputs,
         which means it cannot encounter a run-time error.
+
+        :param timeout: Optional timeout for this request (in seconds).
         """
-        self.most_recent_result = CryptolSafe(self, expr, solver)
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolSafe(self, expr, solver, timeout)
         return self.most_recent_result
 
-    def names(self) -> argo.Command:
-        """Discover the list of names currently in scope in the current context."""
-        self.most_recent_result = CryptolNames(self)
+    def names(self, *, timeout:Optional[float] = None) -> argo.Command:
+        """Discover the list of names currently in scope in the current context.
+
+        :param timeout: Optional timeout for this request (in seconds)."""
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolNames(self, timeout)
         return self.most_recent_result
 
-    def focused_module(self) -> argo.Command:
-        """Return the name of the currently-focused module."""
-        self.most_recent_result = CryptolFocusedModule(self)
+    def focused_module(self, *, timeout:Optional[float] = None) -> argo.Command:
+        """Return the name of the currently-focused module.
+
+        :param timeout: Optional timeout for this request (in seconds)."""
+        timeout = timeout if timeout is not None else self.timeout
+        self.most_recent_result = CryptolFocusedModule(self, timeout)
         return self.most_recent_result
 
     def reset(self) -> None:
@@ -258,6 +406,14 @@ class CryptolConnection:
         CryptolResetServer(self)
         self.most_recent_result = None
 
+    def interrupt(self) -> None:
+        """Interrupt the Cryptol server, cancelling any in-progress requests."""
+        CryptolInterrupt(self)
+
+    def logging(self, on : bool, *, dest : TextIO = sys.stderr) -> None:
+        """Whether to log received and transmitted JSON."""
+        self.server_connection.logging(on=on,dest=dest)
+
 class CryptolDynamicSocketProcess(DynamicSocketProcess):
 
     def __init__(self, command: str, *,
@@ -313,7 +469,7 @@ class CryptolStdIOProcess(StdIOProcess):
 #         current_expr = self.name
 #         found_args = []
 #         while len(arg_types) > 0 and len(remaining_args) > 0:
-#             found_args.append(arg_types[0].from_python(remaining_args[0]))
+#             found_args.append(to_cryptol(remaining_args[0]))
 #             current_expr = {'expression': 'call', 'function': self.name, 'arguments': found_args}
 #             ty = self.connection.check_type(current_expr).result()
 #             current_type = cryptoltypes.to_schema(ty)
